#Original Engine Source:
#Copyright (C) 2013 Michael Fogleman - https://github.com/fogleman/Minecraft

#Image to Sound Sonifying technique:
#Copyright(C)Peter B.L. Meijer - http://www.seeingwithsound.com/im2sound.htm

#Modified by: Magurp244 - 2015 ~ 09/04/18

import six
import packaging
import packaging.version
import packaging.specifiers
import packaging.requirements
from collections import deque
import pyglet
pyglet.options['shadow_window'] = False
from pyglet import image
from pyglet.gl import *
from pyglet.graphics import TextureGroup
from pyglet.window import key, mouse

from pyglet.media.drivers.openal import lib_openal as al
from pyglet.media.drivers.openal import lib_alc as alc
import struct
import numpy
import math
import random
import time
import ctypes
import sys
import os
import wave

import pyttsx

TICKS_PER_SEC = 60

# Size of sectors used to ease block loading.
SECTOR_SIZE = 16

WALKING_SPEED = 5
FLYING_SPEED = 15

GRAVITY = 20.0
MAX_JUMP_HEIGHT = 1.0 # About the height of a block.
# To derive the formula for calculating jump speed, first solve
#    v_t = v_0 + a * t
# for the time at which you achieve maximum height, where a is the acceleration
# due to gravity and v_t = 0. This gives:
#    t = - v_0 / a
# Use t and the desired MAX_JUMP_HEIGHT to solve for v_0 (jump speed) in
#    s = s_0 + v_0 * t + (a * t^2) / 2
JUMP_SPEED = math.sqrt(2 * GRAVITY * MAX_JUMP_HEIGHT)
TERMINAL_VELOCITY = 50

PLAYER_HEIGHT = 2

def cube_vertices(x, y, z, n):
    """ Return the vertices of the cube at position x, y, z with size 2*n.

    """
    return [
        x-n,y+n,z-n, x-n,y+n,z+n, x+n,y+n,z+n, x+n,y+n,z-n,  # top
        x-n,y-n,z-n, x+n,y-n,z-n, x+n,y-n,z+n, x-n,y-n,z+n,  # bottom
        x-n,y-n,z-n, x-n,y-n,z+n, x-n,y+n,z+n, x-n,y+n,z-n,  # left
        x+n,y-n,z+n, x+n,y-n,z-n, x+n,y+n,z-n, x+n,y+n,z+n,  # right
        x-n,y-n,z+n, x+n,y-n,z+n, x+n,y+n,z+n, x-n,y+n,z+n,  # front
        x+n,y-n,z-n, x-n,y-n,z-n, x-n,y+n,z-n, x+n,y+n,z-n,  # back
    ]


def tex_coord(x, y, n=4):
    """ Return the bounding vertices of the texture square.

    """
    m = 1.0 / n
    dx = x * m
    dy = y * m
    return dx, dy, dx + m, dy, dx + m, dy + m, dx, dy + m


def tex_coords(top, bottom, side):
    """ Return a list of the texture squares for the top, bottom and side.

    """
    top = tex_coord(*top)
    bottom = tex_coord(*bottom)
    side = tex_coord(*side)
    result = []
    result.extend(top)
    result.extend(bottom)
    result.extend(side * 4)
    return result


TEXTURE_PATH = 'texture-b.png'

GRASS = tex_coords((1, 0), (0, 1), (0, 0))
SAND = tex_coords((1, 1), (1, 1), (1, 1))
BRICK = tex_coords((2, 0), (2, 0), (2, 0))
STONE = tex_coords((2, 1), (2, 1), (2, 1))

FACES = [
    ( 0, 1, 0),
    ( 0,-1, 0),
    (-1, 0, 0),
    ( 1, 0, 0),
    ( 0, 0, 1),
    ( 0, 0,-1),
]


def normalize(position):
    """ Accepts `position` of arbitrary precision and returns the block
    containing that position.

    Parameters
    ----------
    position : tuple of len 3

    Returns
    -------
    block_position : tuple of ints of len 3

    """
    x, y, z = position
    x, y, z = (int(round(x)), int(round(y)), int(round(z)))
    return (x, y, z)


def sectorize(position):
    """ Returns a tuple representing the sector for the given `position`.

    Parameters
    ----------
    position : tuple of len 3

    Returns
    -------
    sector : tuple of len 3

    """
    x, y, z = normalize(position)
    x, y, z = x / SECTOR_SIZE, y / SECTOR_SIZE, z / SECTOR_SIZE
    return (x, 0, z)


class Model(object):

    def __init__(self):

        # A Batch is a collection of vertex lists for batched rendering.
        self.batch = pyglet.graphics.Batch()

        # A TextureGroup manages an OpenGL texture.
        self.group = TextureGroup(image.load(TEXTURE_PATH).get_texture())

        # A mapping from position to the texture of the block at that position.
        # This defines all the blocks that are currently in the world.
        self.world = {}

        # Same mapping as `world` but only contains blocks that are shown.
        self.shown = {}

        # Mapping from position to a pyglet `VertextList` for all shown blocks.
        self._shown = {}

        # Mapping from sector to a list of positions inside that sector.
        self.sectors = {}

        # Simple function queue implementation. The queue is populated with
        # _show_block() and _hide_block() calls
        self.queue = deque()

        self._initialize()

    def _initialize(self):
        """ Initialize the world by placing all the blocks.

        """
        n = 80  # 1/2 width and height of world
        s = 1  # step size
        y = 0  # initial y height
        for x in xrange(-n, n + 1, s):
            for z in xrange(-n, n + 1, s):
                # create a layer stone an grass everywhere.
                self.add_block((x, y - 2, z), GRASS, immediate=False)
                self.add_block((x, y - 3, z), STONE, immediate=False)
                if x in (-n, n) or z in (-n, n):
                    # create outer walls.
                    for dy in xrange(-2, 3):
                        self.add_block((x, y + dy, z), STONE, immediate=False)

        # generate the hills randomly
        o = n - 10
        for _ in xrange(120):
            a = random.randint(-o, o)  # x position of the hill
            b = random.randint(-o, o)  # z position of the hill
            c = -1  # base of the hill
            h = random.randint(1, 6)  # height of the hill
            s = random.randint(4, 8)  # 2 * s is the side length of the hill
            d = 1  # how quickly to taper off the hills
            t = random.choice([GRASS, SAND, BRICK])
            for y in xrange(c, c + h):
                for x in xrange(a - s, a + s + 1):
                    for z in xrange(b - s, b + s + 1):
                        if (x - a) ** 2 + (z - b) ** 2 > (s + 1) ** 2:
                            continue
                        if (x - 0) ** 2 + (z - 0) ** 2 < 5 ** 2:
                            continue
                        self.add_block((x, y, z), t, immediate=False)
                s -= d  # decrement side lenth so hills taper off

    def hit_test(self, position, vector, max_distance=8):
        """ Line of sight search from current position. If a block is
        intersected it is returned, along with the block previously in the line
        of sight. If no block is found, return None, None.

        Parameters
        ----------
        position : tuple of len 3
            The (x, y, z) position to check visibility from.
        vector : tuple of len 3
            The line of sight vector.
        max_distance : int
            How many blocks away to search for a hit.

        """
        m = 8
        x, y, z = position
        dx, dy, dz = vector
        previous = None
        for _ in xrange(max_distance * m):
            key = normalize((x, y, z))
            if key != previous and key in self.world:
                return key, previous
            previous = key
            x, y, z = x + dx / m, y + dy / m, z + dz / m
        return None, None

    def exposed(self, position):
        """ Returns False is given `position` is surrounded on all 6 sides by
        blocks, True otherwise.

        """
        x, y, z = position
        for dx, dy, dz in FACES:
            if (x + dx, y + dy, z + dz) not in self.world:
                return True
        return False

    def add_block(self, position, texture, immediate=True):
        """ Add a block with the given `texture` and `position` to the world.

        Parameters
        ----------
        position : tuple of len 3
            The (x, y, z) position of the block to add.
        texture : list of len 3
            The coordinates of the texture squares. Use `tex_coords()` to
            generate.
        immediate : bool
            Whether or not to draw the block immediately.

        """
        if position in self.world:
            self.remove_block(position, immediate)
        self.world[position] = texture
        self.sectors.setdefault(sectorize(position), []).append(position)
        if immediate:
            if self.exposed(position):
                self.show_block(position)
            self.check_neighbors(position)

    def remove_block(self, position, immediate=True):
        """ Remove the block at the given `position`.

        Parameters
        ----------
        position : tuple of len 3
            The (x, y, z) position of the block to remove.
        immediate : bool
            Whether or not to immediately remove block from canvas.

        """
        del self.world[position]
        self.sectors[sectorize(position)].remove(position)
        if immediate:
            if position in self.shown:
                self.hide_block(position)
            self.check_neighbors(position)

    def check_neighbors(self, position):
        """ Check all blocks surrounding `position` and ensure their visual
        state is current. This means hiding blocks that are not exposed and
        ensuring that all exposed blocks are shown. Usually used after a block
        is added or removed.

        """
        x, y, z = position
        for dx, dy, dz in FACES:
            key = (x + dx, y + dy, z + dz)
            if key not in self.world:
                continue
            if self.exposed(key):
                if key not in self.shown:
                    self.show_block(key)
            else:
                if key in self.shown:
                    self.hide_block(key)

    def show_block(self, position, immediate=True):
        """ Show the block at the given `position`. This method assumes the
        block has already been added with add_block()

        Parameters
        ----------
        position : tuple of len 3
            The (x, y, z) position of the block to show.
        immediate : bool
            Whether or not to show the block immediately.

        """
        texture = self.world[position]
        self.shown[position] = texture
        if immediate:
            self._show_block(position, texture)
        else:
            self._enqueue(self._show_block, position, texture)

    def _show_block(self, position, texture):
        """ Private implementation of the `show_block()` method.

        Parameters
        ----------
        position : tuple of len 3
            The (x, y, z) position of the block to show.
        texture : list of len 3
            The coordinates of the texture squares. Use `tex_coords()` to
            generate.

        """
        x, y, z = position
        vertex_data = cube_vertices(x, y, z, 0.5)
        texture_data = list(texture)
        # create vertex list 
        # FIXME Maybe `add_indexed()` should be used instead
        self._shown[position] = self.batch.add(24, GL_QUADS, self.group,
            ('v3f/static', vertex_data),
            ('t2f/static', texture_data))

    def hide_block(self, position, immediate=True):
        """ Hide the block at the given `position`. Hiding does not remove the
        block from the world.

        Parameters
        ----------
        position : tuple of len 3
            The (x, y, z) position of the block to hide.
        immediate : bool
            Whether or not to immediately remove the block from the canvas.

        """
        self.shown.pop(position)
        if immediate:
            self._hide_block(position)
        else:
            self._enqueue(self._hide_block, position)

    def _hide_block(self, position):
        """ Private implementation of the 'hide_block()` method.

        """
        self._shown.pop(position).delete()

    def show_sector(self, sector):
        """ Ensure all blocks in the given sector that should be shown are
        drawn to the canvas.

        """
        for position in self.sectors.get(sector, []):
            if position not in self.shown and self.exposed(position):
                self.show_block(position, False)

    def hide_sector(self, sector):
        """ Ensure all blocks in the given sector that should be hidden are
        removed from the canvas.

        """
        for position in self.sectors.get(sector, []):
            if position in self.shown:
                self.hide_block(position, False)

    def change_sectors(self, before, after):
        """ Move from sector `before` to sector `after`. A sector is a
        contiguous x, y sub-region of world. Sectors are used to speed up
        world rendering.

        """
        before_set = set()
        after_set = set()
        pad = 4
        for dx in xrange(-pad, pad + 1):
            for dy in [0]:  # xrange(-pad, pad + 1):
                for dz in xrange(-pad, pad + 1):
                    if dx ** 2 + dy ** 2 + dz ** 2 > (pad + 1) ** 2:
                        continue
                    if before:
                        x, y, z = before
                        before_set.add((x + dx, y + dy, z + dz))
                    if after:
                        x, y, z = after
                        after_set.add((x + dx, y + dy, z + dz))
        show = after_set - before_set
        hide = before_set - after_set
        for sector in show:
            self.show_sector(sector)
        for sector in hide:
            self.hide_sector(sector)

    def _enqueue(self, func, *args):
        """ Add `func` to the internal queue.

        """
        self.queue.append((func, args))

    def _dequeue(self):
        """ Pop the top function from the internal queue and call it.

        """
        func, args = self.queue.popleft()
        func(*args)

    def process_queue(self):
        """ Process the entire queue while taking periodic breaks. This allows
        the game loop to run smoothly. The queue contains calls to
        _show_block() and _hide_block() so this method should be called if
        add_block() or remove_block() was called with immediate=False

        """
        start = time.clock()
        while self.queue and time.clock() - start < 1.0 / TICKS_PER_SEC:
            self._dequeue()

    def process_entire_queue(self):
        """ Process the entire queue with no breaks.

        """
        while self.queue:
            self._dequeue()


class Window(pyglet.window.Window):

    def __init__(self, *args, **kwargs):
        super(Window, self).__init__(*args, **kwargs)

        # Whether or not the window exclusively captures the mouse.
        self.exclusive = False

        # When flying gravity has no effect and speed is increased.
        self.flying = False

        # Strafing is moving lateral to the direction you are facing,
        # e.g. moving to the left or right while continuing to face forward.
        #
        # First element is -1 when moving forward, 1 when moving back, and 0
        # otherwise. The second element is -1 when moving left, 1 when moving
        # right, and 0 otherwise.
        self.strafe = [0, 0]

        # Current (x, y, z) position in the world, specified with floats. Note
        # that, perhaps unlike in math class, the y-axis is the vertical axis.
        self.position = [0, 0, 0]

        # First element is rotation of the player in the x-z plane (ground
        # plane) measured from the z-axis down. The second is the rotation
        # angle from the ground plane up. Rotation is in degrees.
        #
        # The vertical plane rotation ranges from -90 (looking straight down) to
        # 90 (looking straight up). The horizontal rotation range is unbounded.
        self.rotation = [0, 0]

        # Which sector the player is currently in.
        self.sector = None

        # The crosshairs at the center of the screen.
        self.reticle = None

        # Velocity in the y (upward) direction.
        self.dy = 0

        # A list of blocks the player can place. Hit num keys to cycle.
        self.inventory = [BRICK, GRASS, SAND]

        # The current block the user can place. Hit num keys to cycle.
        self.block = self.inventory[0]

        # Convenience list of num keys.
        self.num_keys = [
            key._1, key._2, key._3, key._4, key._5,
            key._6, key._7, key._8, key._9, key._0]

        # Instance of the model that handles the world.
        self.model = Model()

        # The label that is displayed in the top left of the canvas.
        self.label = pyglet.text.Label('', font_name='Arial', font_size=18,
            x=10, y=self.height - 10, anchor_x='left', anchor_y='top',
            color=(0, 0, 0, 255))

    #load listener
        self.listener = Listener()
        self.listener.position = [320,240,0]
    #load ambient movement sounds
        self.source_pool = [load_sound('walk2.wav'), #walk tone 0
                            load_sound('fall2.wav'), #fall tone 1
                            load_sound('hit3.wav'), #hit tone 2
                            load_sound('jump3.wav'), #jump tone 3
                            load_sound('remove.wav'), #destroy block 4
                            load_sound('place.wav'), #place block 5
                            load_sound('tick3.wav'), #horizontal look 6
                            load_sound('tock.wav'), #vertical look 7
                            load_sound('brick3.wav'), #brick 8
                            load_sound('sand3.wav'), #sand 9
                            load_sound('dirt2.wav'), #grass 10
                            load_sound('stone3.wav'), #stone 11
                            load_sound('vsand.wav'), #voice sand 12
                            load_sound('vdirt.wav'), #voice dirt 13
                            load_sound('vbrick.wav'), #voice brick 14
                            load_sound('vstone.wav')] #voice stone 15

    #destroy block sound player
        self.d_block = Player()
        self.d_block.add(self.source_pool[4])
        self.d_block.position = [320,240,0]
    #place block sound player
        self.p_block = Player()
        self.p_block.add(self.source_pool[5])
        self.p_block.position = [320,240,0]
    #vertical/horrizontal sound players
        self.v_scroll = Player()
        self.v_scroll.add(self.source_pool[7])
        self.v_scroll.position = [320,240,0]
        self.h_scroll = Player()
        self.h_scroll.add(self.source_pool[6])

    #block sound and ID voice player
        self.b_sound = Player()
        self.b_trigger = False
        self.v_trigger = False
    #walking sound player
        self.walk = Player()
        self.walk.add(self.source_pool[0])
        self.walk.position = [320,0,0] #alternate a bit left/right for foot simulation?
    #falling sound player
        self.fall = Player()
        self.fall.add(self.source_pool[1])
        self.fall.position = [320,0,0]
    #collision sound player
        self.hit = Player()
        self.hit.add(self.source_pool[2])
        self.hit.position = [320,240,0] #alternate position based on impact?
    #jumping sound pkayer
        self.jump = Player()
        self.jump.add(self.source_pool[3])
        self.jump.position = [320,0,0]

    #Meijer screen scan player
        self.scan = Player()
        self.scan_gen = sonifier()#sweep()
        self.scan_toggle = False
        self.scan_buffer = buffer_sound()
        self.scan_loop = False
        self.buffer = None
        self.image_buf = None

    #x/y coords for mouse motion
        self.r_move = [0,0]

        self.sound_pool = [self.walk,self.fall,self.hit,self.jump,
                           self.scan, self.b_sound, self.v_scroll,
                           self.h_scroll, self.d_block,self.p_block]

    #rotation for incremental audio ticks
        self.rx = 0
        self.ry = 0
    #angle points for horrizontal and vertical compass positions
        self.rot = [[0,22.5,45,67.5,90,112.5,135,157.5,180,202.5,225,247.5,270,292.5,315,337.5,360],
                    [0,11.25,22.5,33.75,45,56.25,67.5,78.75,90]]

        glEnable(GL_DEPTH_TEST)

        glEnable(GL_CULL_FACE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)

    #initialize the Text to Speech engine
        self.tts = pyttsx.init()
    #get list of available system voices
        self.voices = {'codex':self.tts.getProperty('voices'),'current':0}
    #set speech speed
        self.tts.setProperty('rate',200)
    #disable tts loop to use our own loop
        self.tts.startLoop(False)


        self.dtrigger = False
        # This call schedules the `update()` method to be called
        # TICKS_PER_SEC. This is the main game event loop.
        pyglet.clock.schedule_interval(self.update, 1.0 / TICKS_PER_SEC)

    def set_exclusive_mouse(self, exclusive):
        """ If `exclusive` is True, the game will capture the mouse, if False
        the game will ignore the mouse.

        """
        super(Window, self).set_exclusive_mouse(exclusive)
        self.exclusive = exclusive

    def get_sight_vector(self):
        """ Returns the current line of sight vector indicating the direction
        the player is looking.

        """
        x, y = self.rotation
        # y ranges from -90 to 90, or -pi/2 to pi/2, so m ranges from 0 to 1 and
        # is 1 when looking ahead parallel to the ground and 0 when looking
        # straight up or down.
        m = math.cos(math.radians(y))
        # dy ranges from -1 to 1 and is -1 when looking straight down and 1 when
        # looking straight up.
        dy = math.sin(math.radians(y))
        dx = math.cos(math.radians(x - 90)) * m
        dz = math.sin(math.radians(x - 90)) * m
        return (dx, dy, dz)

    def get_motion_vector(self):
        """ Returns the current motion vector indicating the velocity of the
        player.

        Returns
        -------
        vector : tuple of len 3
            Tuple containing the velocity in x, y, and z respectively.

        """
        if any(self.strafe):
            x, y = self.rotation
            strafe = math.degrees(math.atan2(*self.strafe))
            y_angle = math.radians(y)
            x_angle = math.radians(x + strafe)
            if self.flying:
                m = math.cos(y_angle)
                dy = math.sin(y_angle)
                if self.strafe[1]:
                    # Moving left or right.
                    dy = 0.0
                    m = 1
                if self.strafe[0] > 0:
                    # Moving backwards.
                    dy *= -1
                # When you are flying up or down, you have less left and right
                # motion.
                dx = math.cos(x_angle) * m
                dz = math.sin(x_angle) * m
            else:
                dy = 0.0
                dx = math.cos(x_angle)
                dz = math.sin(x_angle)
        else:
            dy = 0.0
            dx = 0.0
            dz = 0.0
        return (dx, dy, dz)

    def update(self, dt):
        """ This method is scheduled to be called repeatedly by the pyglet
        clock.

        Parameters
        ----------
        dt : float
            The change in time since the last call.

        """
        self.model.process_queue()
        sector = sectorize(self.position)
        if sector != self.sector:
            self.model.change_sectors(self.sector, sector)
            if self.sector is None:
                self.model.process_entire_queue()
            self.sector = sector
        m = 8
        dt = min(dt, 0.2)
        for _ in xrange(m):
            self._update(dt / m)

     #update tts engine
        self.tts.iterate()


    #process image
        if self.buffer != None and self.scan_gen.output == None:
            self.scan_gen.generate(self.image_buf, self.buffer)
            self.buffer = None

    #purge/load new sound data
        elif self.scan.playing() == False and self.scan_gen.output != None:
            if len(self.scan.queue) > 0:
                self.scan.remove()
            self.scan_buffer.load(self.scan_gen.output)
            self.scan.add(self.scan_buffer)
            self.scan.play()
            self.scan_gen.output = None
            if self.scan_loop == True:
                self.dtrigger = True

    #Move player position for stereo panning effect
        if self.scan.playing():
            self.scan.position = [-100+(self.scan.seek*100),240,50]

    #window scrolling audio code
        for a in xrange(0,len(self.rot[0]),1):
        #scroll right
            if self.rotation[0] >= 0:
            #clockwise movement
                if self.r_move[0] < self.rot[0][a] and self.rotation[0] >= self.rot[0][a]:
                    self.rx = a
                    self.h_scroll.position = [640,240,0]
                    if self.rot[0][a] <= 180:
                        self.h_scroll.pitch = (0.0625 * a)+0.5
                    else:
                        self.h_scroll.pitch = 1.0 - (0.0625 * (a-8))
                    self.h_scroll.play()
            #clockwise reverse movement
                elif self.r_move[0] > self.rot[0][a] and self.rotation[0] <= self.rot[0][a]:
                    self.rx = a
                    self.h_scroll.position = [0,240,0]
                    if self.rot[0][a] <= 180:
                        self.h_scroll.pitch = (0.0625 * a)+0.5
                    else:
                        self.h_scroll.pitch = 1.0 - (0.0625 * (a-8))
                    self.h_scroll.play()
            #roll over to 0
                elif self.r_move[0] > 359 and self.rotation[0] < 1:
                    self.rx = 0
                    self.h_scroll.position = [640,240,0]
                    self.h_scroll.pitch = 0.5
                    self.h_scroll.play()
        #scroll left
            elif self.rotation[0] < 0:
            #counter-clockwise movement
                if self.r_move[0] > -self.rot[0][a] and self.rotation[0] <= -self.rot[0][a]:
                    self.rx = -a
                    self.h_scroll.position = [0,240,0]
                    if self.rot[0][a] <= 180:
                        self.h_scroll.pitch = (0.0625 * a)+0.5
                    else:
                        self.h_scroll.pitch = 1.0 - (0.0625 * (a-8))
                    self.h_scroll.play()
            #counter-clockwise reverse movement
                elif self.r_move[0] < -self.rot[0][a] and self.rotation[0] >= -self.rot[0][a]:
                    self.rx = -a
                    self.h_scroll.position = [640,240,0]
                    if self.rot[0][a] <= 180:
                        self.h_scroll.pitch = (0.0625 * a)+0.5
                    else:
                        self.h_scroll.pitch = 1.0 - (0.0625 * (a-8))
                    self.h_scroll.play()
            #roll over to 0
                elif self.r_move[0] < -359 and self.rotation[0] > -1:
                    self.rx = 0
                    self.h_scroll.position = [0,240,0]
                    self.h_scroll.pitch = 0.5
                    self.h_scroll.play()


        for a in xrange(0,9,1):
            if self.rotation[1] >= 0:
            #scrolling up
                if self.r_move[1] < self.rot[1][a] and self.rotation[1] >= self.rot[1][a]:
                    self.ry = a
                    self.v_scroll.pitch = (0.0625 * a)+1.0
                    self.v_scroll.play()
            #scrolling down
                elif self.r_move[1] > self.rot[1][a] and self.rotation[1] <= self.rot[1][a]:
                    self.ry = a
                    self.v_scroll.pitch = (0.0625 * a)+1.0
                    self.v_scroll.play()
            #scroll to very top
                elif self.r_move[1] < 90 and self.rotation[1] >= 90:
                    self.ry = 8
                    self.v_scroll.pitch = 1.5
                    self.v_scroll.play()

            if self.rotation[1] < 0:
            #scroll down
                if self.r_move[1] > -self.rot[1][a] and self.rotation[1] <= -self.rot[1][a]:
                    self.ry = -a
                    self.v_scroll.pitch = 1.0-(0.0625 * a)
                    self.v_scroll.play()
            #scroll up
                elif self.r_move[1] < -self.rot[1][a] and self.rotation[1] >=  -self.rot[1][a]:
                    self.ry = -a
                    self.v_scroll.pitch = 1.0-(0.0625 * a)
                    self.v_scroll.play()
            #scroll to very bottom
                elif self.r_move[1] > -90 and self.rotation[1] <= -90:
                    self.ry = -8
                    self.v_scroll.pitch = 0.5
                    self.v_scroll.play()

        self.r_move[0] = self.rotation[0]
        self.r_move[1] = self.rotation[1]

    #purge b_sound when finished
        if len(self.b_sound.queue) > 0 and self.b_sound.playing() == False:
            self.b_sound.remove()


    def _update(self, dt):
        """ Private implementation of the `update()` method. This is where most
        of the motion logic lives, along with gravity and collision detection.

        Parameters
        ----------
        dt : float
            The change in time since the last call.

        """
        # walking
        speed = FLYING_SPEED if self.flying else WALKING_SPEED
        d = dt * speed # distance covered this tick.
        dx, dy, dz = self.get_motion_vector()
        # New position in space, before accounting for gravity.
        dx, dy, dz = dx * d, dy * d, dz * d
        # gravity
        if not self.flying:
            # Update your vertical speed: if you are falling, speed up until you
            # hit terminal velocity; if you are jumping, slow down until you
            # start falling.
            self.dy -= dt * GRAVITY
            self.dy = max(self.dy, -TERMINAL_VELOCITY)
            dy += self.dy * dt
        # collisions
        x, y, z = self.position
        x, y, z = self.collide((x + dx, y + dy, z + dz), PLAYER_HEIGHT)

    #play sound when colliding with object above player
        if y == round(y,2) and y != self.position[1]:
            if self.fall.playing() == False:
                self.fall.play()

    #play sound when colliding along X axis
        if dx != 0 and x == round(x,2):
            if x < 0:
            #collision while moving "up" towards 0
                if x - int(x) == -0.25:
                    x += 0.01
                    print 'push'
                    if self.hit.playing() == False:
                        self.hit.play()
            #collision while moving "down" away from 0
                elif x - int(x) == -0.75:
                    x -= 0.01
                    print 'back'
                    if self.hit.playing() == False:
                        self.hit.play()
            else:
            #collition while moving "up" from 0
                if x - int(x) == 0.25:
                    x -= 0.01
                    print 'push'
                    if self.hit.playing() == False:
                        self.hit.play()
            #collision while moving "down" to 0
                elif x - int(x) == 0.75:
                    x += 0.01
                    print 'back'
                    if self.hit.playing() == False:
                        self.hit.play()


    #play sound when colliding along Z axis
        if dz != 0 and z == round(z,2):
            if z < 0:
            #collision while moving "up" towards 0
                if z - int(z) == -0.25:
                    z += 0.01
                    print 'z push'
                    if self.hit.playing() == False:
                        self.hit.play()
            #collision while moving "down" away from 0
                elif z - int(z) == -0.75:
                    z -= 0.01
                    print 'z back'
                    if self.hit.playing() == False:
                        self.hit.play()
            else:
            #collition while moving "up" from 0
                if z - int(z) == 0.25:
                    z -= 0.01
                    print 'z push'
                    if self.hit.playing() == False:
                        self.hit.play()
            #collision while moving "down" to 0
                elif z - int(z) == 0.75:
                    z += 0.01
                    print 'z back'
                    if self.hit.playing() == False:
                        self.hit.play()

    #play sound when walking (but not while jumping/falling)
        if (self.position[0] != x or self.position[2] != z) and self.position[1] == y and self.walk.playing() != True:
            self.walk.play()
            
        self.position = [x, y, z]

    def collide(self, position, height):
        """ Checks to see if the player at the given `position` and `height`
        is colliding with any blocks in the world.

        Parameters
        ----------
        position : tuple of len 3
            The (x, y, z) position to check for collisions at.
        height : int or float
            The height of the player.

        Returns
        -------
        position : tuple of len 3
            The new position of the player taking into account collisions.

        """
        # How much overlap with a dimension of a surrounding block you need to
        # have to count as a collision. If 0, touching terrain at all counts as
        # a collision. If .49, you sink into the ground, as if walking through
        # tall grass. If >= .5, you'll fall through the ground.
        pad = 0#-0.1#0.25
        p = list(position)
        np = normalize(position)
        for face in FACES:  # check all surrounding blocks
            for i in xrange(3):  # check each dimension independently
                if not face[i]:
                    continue
                # How much overlap you have with this dimension.
                d = (p[i] - np[i]) * face[i]
                if d < pad:
                    continue
                for dy in xrange(height):  # check each height
                    op = list(np)
                    op[1] -= dy
                    op[i] += face[i]
                    if tuple(op) not in self.model.world:
                        continue
                    p[i] -= (d - pad) * face[i]
                    if face == (0, -1, 0) or face == (0, 1, 0):
                        # You are colliding with the ground or ceiling, so stop
                        # falling / rising.
                        self.dy = 0
                    break
        return list(p)

    def on_mouse_press(self, x, y, button, modifiers):
        """ Called when a mouse button is pressed. See pyglet docs for button
        amd modifier mappings.

        Parameters
        ----------
        x, y : int
            The coordinates of the mouse click. Always center of the screen if
            the mouse is captured.
        button : int
            Number representing mouse button that was clicked. 1 = left button,
            4 = right button.
        modifiers : int
            Number representing any modifying keys that were pressed when the
            mouse button was clicked.

        """
        if self.exclusive:
            vector = self.get_sight_vector()
            block, previous = self.model.hit_test(self.position, vector)
            if (button == mouse.RIGHT) or \
                    ((button == mouse.LEFT) and (modifiers & key.MOD_CTRL)):
                # ON OSX, control + left click = right click.
                if previous:
                    self.model.add_block(previous, self.block)
                    if self.p_block.playing() == True:
                        self.p_block.stop()
                    self.p_block.play()
            elif button == pyglet.window.mouse.LEFT and block:
                texture = self.model.world[block]
                if texture != STONE:
                    self.model.remove_block(block)
                    if self.d_block.playing() == True:
                        self.d_block.stop()
                    self.d_block.play()
        else:
            self.set_exclusive_mouse(True)

    def on_mouse_motion(self, x, y, dx, dy):
        """ Called when the player moves the mouse.

        Parameters
        ----------
        x, y : int
            The coordinates of the mouse click. Always center of the screen if
            the mouse is captured.
        dx, dy : float
            The movement of the mouse.

        """
        if self.exclusive:
            m = 0.15
            x, y = self.rotation
            x, y = x + dx * m, y + dy * m
            y = max(-90, min(90, y))
            if x < -360:
                x += 360
            elif x > 360:
                x -= 360
            self.rotation = [x, y]

    def on_key_press(self, symbol, modifiers):
        """ Called when the player presses a key. See pyglet docs for key
        mappings.

        Parameters
        ----------
        symbol : int
            Number representing the key that was pressed.
        modifiers : int
            Number representing any modifying keys that were pressed.

        """
        if symbol == key.W:
            self.strafe[0] -= 1
        elif symbol == key.S:
            self.strafe[0] += 1
        elif symbol == key.A:
            self.strafe[1] -= 1
        elif symbol == key.D:
            self.strafe[1] += 1
        elif symbol == key.SPACE:
            if self.dy == 0:
                self.jump.stop()
                self.jump.play()
                self.dy = JUMP_SPEED
    #close program
        elif symbol == key.ESCAPE:
            for a in xrange(0,len(self.sound_pool),1):
                self.sound_pool[a].stop()
                self.sound_pool[a].remove()
                self.sound_pool[a].delete()
            for a in xrange(0,len(self.source_pool),1):
                self.source_pool[a].delete()
            self.scan_buffer.delete()
            self.listener.delete()
            self.tts.endLoop()
            self.close()
    #fly mode
        elif symbol == key.TAB:
            self.flying = not self.flying
        elif symbol in self.num_keys:
            index = (symbol - self.num_keys[0]) % len(self.inventory)
            self.block = self.inventory[index]
            if index == 0:
                if self.tts.isBusy() == True:
                    self.tts.stop()
                self.tts.say('brick selected')
            elif index == 1:
                if self.tts.isBusy() == True:
                    self.tts.stop()
                self.tts.say('grass selected')
            elif index == 2:
                if self.tts.isBusy() == True:
                    self.tts.stop()
                self.tts.say('sand selected')

          

#NOTE: get facing direction, floor/ceil based on -/+ of coords

    #toggle block sound ID
        if symbol == key.Q:
            self.b_trigger = True
            print 'position',self.position

    #toggle block voice ID
        if symbol == key.Z:
            self.v_trigger = True

        vector = self.get_sight_vector()
        block, previous = self.model.hit_test(self.position, vector)
    #place block
        if symbol == key.RETURN:
            if previous:
                self.model.add_block(previous, self.block)
                if self.p_block.playing() == True:
                    self.p_block.stop()
                self.p_block.play()
    #remove block
        if symbol == key.LSHIFT and block:
            texture = self.model.world[block]
            if texture != STONE:
                self.model.remove_block(block)
                if self.d_block.playing() == True:
                    self.d_block.stop()
                self.d_block.play()


    #scroll window up
        if symbol == key.UP:
            if self.ry < 8:
                if self.ry >= 0:
                    self.rotation[1] = self.rot[1][self.ry+1]
                else:
                    self.rotation[1] = -self.rot[1][-(self.ry+1)]
    #scroll window down
        elif symbol == key.DOWN:
            if self.ry > -8:
                if self.ry <= 0:
                    self.rotation[1] = -self.rot[1][-(self.ry-1)]
                else:
                    self.rotation[1] = self.rot[1][self.ry-1]
        print self.rotation[1], self.ry
    #scroll window right
        if symbol == key.RIGHT:
            if self.rx < 16:
                if self.rx >= 0:
                    self.rotation[0] = self.rot[0][self.rx+1]
                else:
                    self.rotation[0] = -self.rot[0][-(self.rx+1)]
            if self.rx + 1 == 16:
                self.rotation[0] = 0
                self.r_move[0] = 360
    #scroll window left
        elif symbol == key.LEFT:
            if self.rx > -16:
                if self.rx <= 0:
                    self.rotation[0] = -self.rot[0][-(self.rx-1)]
                else:
                    self.rotation[0] = self.rot[0][self.rx-1]
            if self.rx - 1 == -16:
                self.rotation[0] = 0
                self.r_move[0] = -360

    #filter 1
        elif symbol == key.F1:
            self.scan_gen.filter = 1 
            if self.tts.isBusy() == True:
                self.tts.stop()
            self.tts.say('Filter dirt')

    #filter 2
        elif symbol == key.F2:
            self.scan_gen.filter = 2
            if self.tts.isBusy() == True:
                self.tts.stop()
            self.tts.say('filter sand')

    #filter 3
        elif symbol == key.F3:
            self.scan_gen.filter = 3
            if self.tts.isBusy() == True:
                self.tts.stop()
            self.tts.say('filter brick')

    #filter 4
        elif symbol == key.F4:
            self.scan_gen.filter = 4 
            if self.tts.isBusy() == True:
                self.tts.stop()
            self.tts.say('filter stone')

    #no filter
        elif symbol == key.F5:
            self.scan_gen.filter = 0
            if self.tts.isBusy() == True:
                self.tts.stop()
            self.tts.say('no filter')

    #loop
        elif symbol == key.F6:
            if self.scan_loop == False:
                self.scan_loop = True
                if self.tts.isBusy() == True:
                    self.tts.stop()
                self.tts.say('loop enabled')

            else:
                self.scan_loop = False
                if self.tts.isBusy() == True:
                    self.tts.stop()
                self.tts.say('loop disabled')

    #speed up
        elif symbol == key.F7:
            self.scan_gen.rate(speed=False)
            if self.tts.isBusy() == True:
                self.tts.stop()
            self.tts.say('speed up')
            print self.scan_gen.speed
    #speed down
        elif symbol == key.F8:
            self.scan_gen.rate(speed=True)
            if self.tts.isBusy() == True:
                self.tts.stop()
            self.tts.say('speed down')
            print self.scan_gen.speed




    def on_key_release(self, symbol, modifiers):
        """ Called when the player releases a key. See pyglet docs for key
        mappings.

        Parameters
        ----------
        symbol : int
            Number representing the key that was pressed.
        modifiers : int
            Number representing any modifying keys that were pressed.

        """
        if symbol == key.W:
            self.strafe[0] += 1
        elif symbol == key.S:
            self.strafe[0] -= 1
        elif symbol == key.A:
            self.strafe[1] += 1
        elif symbol == key.D:
            self.strafe[1] -= 1
        if symbol == key.E:
            self.dtrigger = True


    def on_resize(self, width, height):
        """ Called when the window is resized to a new `width` and `height`.

        """
        # label
        self.label.y = height - 10
        # reticle
        if self.reticle:
            self.reticle.delete()
        x, y = self.width / 2, self.height / 2
        n = 10
        self.reticle = pyglet.graphics.vertex_list(4,
            ('v2i', (x - n, y, x + n, y, x, y - n, x, y + n))
        )

    def set_2d(self):
        """ Configure OpenGL to draw in 2d.

        """
        width, height = self.get_size()
        glDisable(GL_DEPTH_TEST)
        glViewport(0, 0, width, height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, width, 0, height, -1, 1)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

    def set_3d(self):
        """ Configure OpenGL to draw in 3d.

        """
        width, height = self.get_size()
        glViewport(0, 0, width, height)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(65.0, width / float(height), 0.5, 20.0)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        x, y = self.rotation
        glRotatef(x, 0, 1, 0)
        glRotatef(-y, math.cos(math.radians(x)), 0, math.sin(math.radians(x)))
        x, y, z = self.position
        glTranslatef(-x, -y, -z)
        glClear(GL_DEPTH_BUFFER_BIT)
        glEnable(GL_DEPTH_TEST)

    def set_3da(self):
        """ Configure OpenGL to draw in 3d.

        """
        width = 64
        height = 48
        glViewport(0, 0, width, height)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(65.0, width / float(height), 0.5, 20.0)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        x, y = self.rotation
        glRotatef(x, 0, 1, 0)
        glRotatef(-y, math.cos(math.radians(x)), 0, math.sin(math.radians(x)))
        x, y, z = self.position
        glTranslatef(-x, -y, -z)
        glClear(GL_DEPTH_BUFFER_BIT)
        glEnable(GL_DEPTH_TEST)

    def on_draw(self):
        """ Called by pyglet to draw the canvas.

        """
        self.clear()
    #capture 64x64 image for image to sound rendering
        if self.dtrigger == True:
            self.set_3da()
            glPushMatrix()
            glColor3d(1, 1, 1)
            self.model.batch.draw()
            self.draw_focused_block()
            glPopMatrix()
            glDisable(GL_DEPTH_TEST)
            glColor4f(1, 1, 1, 1)
            self.buffer = pyglet.image.get_buffer_manager().get_depth_buffer().get_region(0,0,64,48).get_image_data().get_texture()
            self.image_buf = pyglet.image.get_buffer_manager().get_color_buffer().get_region(0,0,64,48).get_image_data().get_texture()
            glDisable(GL_DEPTH_TEST)
            self.clear()

            self.dtrigger = False
        if self.scan_gen.filter == 0:
            self.set_3da()
            glColor3d(1, 1, 1)
            self.model.batch.draw()
            self.draw_focused_block()
            glDisable(GL_DEPTH_TEST)
            gl.glDepthRange(-1.0,0.7)

            glDepthFunc(GL_LEQUAL)
            glColor4f(1, 1, 1, 1)
            img = pyglet.image.get_buffer_manager().get_depth_buffer().get_region(0,0,64,48).get_image_data()
            glDisable(GL_DEPTH_TEST)

            tmp = img.get_data('L',img.width)
            tmp2 = numpy.fromstring(tmp,dtype='uint8')
            tmp2 = numpy.reshape(tmp2,(64,48,1))
            tmp2 = numpy.invert(tmp2,dtype='uint8')

            img.set_data('L',img.width,tmp2.tobytes())
            img = img.get_texture()

        #disable anti-aliasing/bilinear-filtering
            glTexParameteri(img.target, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
            glTexParameteri(img.target, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
            img.width = 640
            img.height = 480
            self.set_2d()
            img.blit(0,0)

        else:
            self.set_3d()
            glColor3d(1, 1, 1)
            self.model.batch.draw()
            self.draw_focused_block()

    def draw_focused_block(self):
        """ Draw black edges around the block that is currently under the
        crosshairs.

        """
        vector = self.get_sight_vector()
        block = self.model.hit_test(self.position, vector)[0]
        if block:
        #ID particular type of block with tone.
            if self.b_trigger == True:
                if self.b_sound.playing() == True:
                    self.b_sound.stop()
                    self.b_sound.remove()
            #grass/dirt
                if self.model.world[block] == GRASS:
                    self.b_sound.add(self.source_pool[10])
                    if (8.5 - calculateDistance(block[0],block[2],self.position[0],self.position[2]))/8.5 > 0.0:
                        self.b_sound.volume = (8.5 - calculateDistance(block[0],block[2],self.position[0],self.position[2]))/8.5
                    else:
                        self.b_sound.volume = 0.0
                    self.b_sound.play()
            #sand
                elif self.model.world[block] == SAND:
                    self.b_sound.add(self.source_pool[9])
                    if (8.5 - calculateDistance(block[0],block[2],self.position[0],self.position[2]))/8.5 > 0.0:
                        self.b_sound.volume = (8.5 - calculateDistance(block[0],block[2],self.position[0],self.position[2]))/8.5
                    else:
                        self.b_sound.volume = 0.0
                    self.b_sound.play()
            #brick
                elif self.model.world[block] == BRICK:
                    self.b_sound.add(self.source_pool[8])
                    if (8.5 - calculateDistance(block[0],block[2],self.position[0],self.position[2]))/8.5 > 0.0:
                        self.b_sound.volume = (8.5 - calculateDistance(block[0],block[2],self.position[0],self.position[2]))/8.5
                    else:
                        self.b_sound.volume = 0.0
                    self.b_sound.play()
            #stone
                elif self.model.world[block] == STONE:
                    self.b_sound.add(self.source_pool[11])
                    if (8.5 - calculateDistance(block[0],block[2],self.position[0],self.position[2]))/8.5 > 0.0:
                        self.b_sound.volume = (8.5 - calculateDistance(block[0],block[2],self.position[0],self.position[2]))/8.5
                    else:
                        self.b_sound.volume = 0.0
                    self.b_sound.play()
                self.b_trigger = False
        #play voice ID for block type
            elif self.v_trigger == True:
                self.b_sound.volume = 1.0
                if self.b_sound.playing() == True:
                    self.b_sound.stop()
                    self.b_sound.remove()
            #grass/dirt
                if self.model.world[block] == GRASS:
                    self.b_sound.add(self.source_pool[13])
                    self.b_sound.play()
            #sand
                elif self.model.world[block] == SAND:
                    self.b_sound.add(self.source_pool[12])
                    self.b_sound.play()
            #brick
                elif self.model.world[block] == BRICK:
                    self.b_sound.add(self.source_pool[14])
                    self.b_sound.play()
            #stone
                elif self.model.world[block] == STONE:
                    self.b_sound.add(self.source_pool[15])
                    self.b_sound.play()
                self.v_trigger = False

        #draw highlighted block lines
            x, y, z = block
            vertex_data = cube_vertices(x, y, z, 0.51)
            glColor3d(0, 0, 0)
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
            pyglet.graphics.draw(24, GL_QUADS, ('v3f/static', vertex_data))
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)

    def draw_label(self):
        """ Draw the label in the top left of the screen.

        """
        x, y, z = self.position
        self.label.text = '%02d (%.2f, %.2f, %.2f) %d / %d' % (
            pyglet.clock.get_fps(), x, y, z,
            len(self.model._shown), len(self.model.world))
        self.label.draw()

    def draw_reticle(self):
        """ Draw the crosshairs in the center of the screen.

        """
        pass #causes memory access error
##        glColor3d(0, 0, 0)
##        self.reticle.draw(GL_LINES)


def setup_fog():
    """ Configure the OpenGL fog properties.

    """
    # Enable fog. Fog "blends a fog color with each rasterized pixel fragment's
    # post-texturing color."
    glEnable(GL_FOG)
    # Set the fog color.
    glFogfv(GL_FOG_COLOR, (GLfloat * 4)(0.0, 0.0, 0.0, 1))
    # Say we have no preference between rendering speed and quality.
    glHint(GL_FOG_HINT, GL_DONT_CARE)
    # Specify the equation used to compute the blending factor.
    glFogi(GL_FOG_MODE, GL_LINEAR)
    # How close and far away fog starts and ends. The closer the start and end,
    # the denser the fog in the fog range.
    glFogf(GL_FOG_START, 0.0)
    glFogf(GL_FOG_END, 20.0)


def setup():
    """ Basic OpenGL configuration.

    """
    # Set the color of "clear", i.e. the sky, in rgba.
    glClearColor(0.0, 0.0, 0.0, 1)
    # Enable culling (not rendering) of back-facing facets -- facets that aren't
    # visible to you.
    glEnable(GL_CULL_FACE)
    # Set the texture minification/magnification function to GL_NEAREST (nearest
    # in Manhattan distance) to the specified texture coordinates. GL_NEAREST
    # "is generally faster than GL_LINEAR, but it can produce textured images
    # with sharper edges because the transition between texture elements is not
    # as smooth."
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
##    setup_fog()


def main():
    config = Config(sample_buffers=1, samples=4, depth_size=32, double_buffer=True,buffer_size=32)
    window = Window(width=640, height=480, caption='Pyglet', resizable=True,config=config)
    # Hide the mouse cursor and prevent the mouse from leaving the window.
    window.set_exclusive_mouse(True)
    setup()
    pyglet.app.run()


#used to calculate distance between player and block
def calculateDistance(x1,y1,x2,y2):
    return math.sqrt((x2 - x1)**2 + (y2 - y1)**2)



#image to sound converter class
class sonifier(object):
#pre-calculate waveform values
    def __init__(self):
        self.speed = 1.0
    #sample rate/length of our audio data
        self.byte = int(21000 * self.speed)
    #size of target image(s)
        self.x = 64
        self.y = 48
    #adjust frequencies to maintain pitch when adjusting speed
        self.max = int(5000*self.speed)
        self.min = int(500*self.speed)

        self.ns = self.byte
        self.scale = 0.5/math.sqrt(self.x)
        self.TwoPi = 6.283185307179586476925287
        self.rnd = 0.211321159122
        self.phi0 = self.TwoPi * self.rnd

    #Generate sine wave to multiply image data against
        self.w = numpy.asarray(range(self.y))
        self.w = self.TwoPi * self.min * numpy.power(1.0*self.max/self.min, 1.0*self.w/(self.x-1))
        self.w = numpy.tile(self.w,(self.ns,1))

        self.core = numpy.asarray(range(self.ns))
        self.core = numpy.reshape(self.core,(self.byte,1))
        self.core = numpy.repeat(self.core,(self.y),axis=1)
    #finished sine wave
        self.w = numpy.sin(self.w * (self.core * (1.0 / self.byte)) + self.phi0)
    #initialize final output ctypes array
        self.output = None

    #filter index
        self.index = [None,[128,0,0,255],[0,128,0,255],[0,64,128,255],[128,128,128,255]]
        self.index = numpy.asarray(self.index)

    #current filter
        self.filter = 0


#change the dimensions the sonifier should process
    def size(self,width,height):
        self.x = width
        self.y = height

    #sample rate/length of our audio data
        self.byte = int(21000 * self.speed)
    #adjust frequencies to maintain pitch when adjusting speed
        self.max = int(5000*self.speed)
        self.min = int(500*self.speed)

        self.ns = self.byte
        self.scale = 0.5/math.sqrt(self.x)

    #Generate sine wave to multiply image data against
        self.w = numpy.asarray(range(self.y))
        self.w = self.TwoPi * self.min * numpy.power(1.0*self.max/self.min, 1.0*self.w/(self.x-1))
        self.w = numpy.tile(self.w,(self.ns,1))

        self.core = numpy.asarray(range(self.ns))
        self.core = numpy.reshape(self.core,(self.byte,1))
        self.core = numpy.repeat(self.core,(self.y),axis=1)
    #finished sine wave
        self.w = numpy.sin(self.w * (self.core * (1.0 / self.byte)) + self.phi0)
    #initialize final output ctypes array
        self.output = None


#alter output data playback speed
    def rate(self,speed):
        if speed == False:
            if self.speed == 2.0:
                self.speed = 1.5
            elif self.speed == 1.5:
                self.speed = 1.0
            elif self.speed == 1.0:
                self.speed = 0.75
            elif self.speed == 0.75:
                self.speed = 0.5
            elif self.speed == 0.5:
                self.speed = 0.25
            elif self.speed == 0.25:
                self.speed = 0.1
            elif self.speed == 0.1:
                self.speed = 0.05
        elif speed == True:
            if self.speed == 0.05:
                self.speed = 0.1
            elif self.speed == 0.1:
                self.speed = 0.25
            elif self.speed == 0.25:
                self.speed = 0.5
            elif self.speed == 0.5:
                self.speed = 0.75
            elif self.speed == 0.75:
                self.speed = 1.0
            elif self.speed == 1.0:
                self.speed = 1.5
            elif self.speed == 1.5:
                self.speed = 2.0

        self.byte = int(21000 * self.speed)
    #adjust frequencies to maintain pitch when adjusting speed
        self.max = int(5000*self.speed)
        self.min = int(500*self.speed)
        self.ns = self.byte
    #recalculate sine wave
        self.w = numpy.asarray(range(self.y))
        self.w = self.TwoPi * self.min * numpy.power(1.0*self.max/self.min, 1.0*self.w/(self.x-1))
        self.w = numpy.tile(self.w,(self.ns,1))

        self.core = numpy.asarray(range(self.ns))
        self.core = numpy.reshape(self.core,(self.byte,1))
        self.core = numpy.repeat(self.core,(self.y),axis=1)
    #finished sine wave
        self.w = numpy.sin(self.w * (self.core * (1.0 / self.byte)) + self.phi0)
    #reinitialize final output ctypes array
        self.output = None


#load and convert image, multiply it by waveform and load into ctypes array for output
    def generate(self,image,depth):
    #process depth map
        rawimage = depth.get_image_data()
        pixels = rawimage.get_data('RGBA',rawimage.width*4)
    #convert image data to numpy array int16
        self.data = numpy.fromstring(pixels, dtype='uint8')
    #reshape array into 4 by (width*height) array, then strip color data except red channel.
        self.data = numpy.reshape(self.data,(len(self.data)/4,4))
    #invert depth map
        self.data = numpy.invert(self.data,dtype='uint8')

    #process image stencil
        rawimage = image.get_image_data()
        pixels = rawimage.get_data('RGBA',rawimage.width*4)
    #convert image data to numpy array int16
        stencil = numpy.fromstring(pixels, dtype='uint8')
        stencil = numpy.reshape(stencil,(len(stencil)/4,4))

    #filter depthmap data using image as stencil
        if self.filter == 0:
            tmp = stencil > [0,0,0,0]
            tmp = numpy.sum(tmp,axis=1)
            tmp = tmp > 1
        else:
            tmp = stencil == self.index[self.filter]
            tmp = numpy.sum(tmp,axis=1)
            tmp = tmp == 4

        test = numpy.zeros((self.data.shape),dtype='uint8')
        test[tmp] = self.data[tmp]
        self.data = test

    #cull unnecessary data
        self.data = self.data[::,:1:]
    #reshape array into X by Y grid to match image X/Y
        self.data = numpy.reshape(self.data,(self.y,self.x)) #Y, X values
    #convert image data to floating point.
        self.data = self.data / 255.0
    #expand image data to match size of wav sample
        self.data = numpy.repeat(self.data,int(numpy.ceil(self.byte / float(self.x))),axis=1) #X value, not Y
        self.data = self.data[::,:self.byte:]

        self.data = numpy.swapaxes(self.data,0,1) 

    #multiply array for volume adjustment
        self.data = numpy.multiply(self.data,self.w) * 1500.0

    #sum array data into 1D array, creating our final waveform
        self.data = numpy.sum(self.data,axis=1)

    #add click tone to beginning
        #self.data[:64] = ((2.0 * self.rnd - 1.0) / self.scale) * 32767.0

    #load data into a ctypes array for dumping into OpenAL buffer
        self.output = (ctypes.c_short * self.ns)()
        for a in xrange(0,len(self.output),1):
            self.output[a] = self.output[a]+int(self.data[a])



#-----------------------------------------------------------------------------------------------------------------------------------------------/
class load_sound(object):
    def __init__(self,filename):
        self.name = filename
    #load/set audio file
        if len (sys.argv) < 2:
            dirname = os.path.dirname(os.path.realpath(__file__))
            dirname += '/data/'
            fname = os.path.join(dirname, filename)
        else:
            fname = sys.argv[1]

        wavefp = wave.open(fname)
        channels = wavefp.getnchannels()
        bitrate = wavefp.getsampwidth() * 8
        samplerate = wavefp.getframerate()
        wavbuf = wavefp.readframes(wavefp.getnframes())
        self.duration = (len(wavbuf) / float(samplerate))/2
        self.length = len(wavbuf)
        formatmap = {
            (1, 8) : al.AL_FORMAT_MONO8,
            (2, 8) : al.AL_FORMAT_STEREO8,
            (1, 16): al.AL_FORMAT_MONO16,
            (2, 16) : al.AL_FORMAT_STEREO16,
        }
        alformat = formatmap[(channels, bitrate)]

        self.buf = al.ALuint(0)
        al.alGenBuffers(1, self.buf)
    #allocate buffer space to: buffer, format, data, len(data), and samplerate
        al.alBufferData(self.buf, alformat, wavbuf, len(wavbuf), samplerate)

    def delete(self):
        al.alDeleteBuffers(1, self.buf)

#-------------------------------------------------------------------------------------------------------|
#OpenAL Sound Buffer
class buffer_sound(object):
    def __init__(self):
        self.channels = 1
        self.bitrate = 16
        self.samplerate = 21000
        self.wavbuf = None
        self.alformat = al.AL_FORMAT_MONO16
        self.length = None
##        formatmap = {
##            (1, 8) : al.AL_FORMAT_MONO8,
##            (2, 8) : al.AL_FORMAT_STEREO8,
##            (1, 16): al.AL_FORMAT_MONO16,
##            (2, 16) : al.AL_FORMAT_STEREO16,
##        }
##        alformat = formatmap[(channels, bitrate)]


        self.buf = al.ALuint(0)
        al.alGenBuffers(1, self.buf)

    def load(self,data):
        self.wavbuf = data
        self.length = len(data)
    #allocate buffer space to: buffer, format, data, len(data), and samplerate
        al.alBufferData(self.buf, self.alformat, self.wavbuf, len(self.wavbuf)*2, self.samplerate)

    def delete(self):
        al.alDeleteBuffers(1, self.buf)


class Listener(object):
    def __init__(self):
    #load device/context/listener
        self.device = alc.alcOpenDevice(None)
        self.context = alc.alcCreateContext(self.device, None)
        alc.alcMakeContextCurrent(self.context)

#set player position
    def _set_position(self,pos):
        self._position = pos
        x,y,z = map(int, pos)
        al.alListener3f(al.AL_POSITION, x, y, z)

    def _get_position(self):
        return self._position

    def delete(self):
        alc.alcDestroyContext(self.context)
        alc.alcCloseDevice(self.device)

    position = property(_get_position, _set_position,doc="""get/set position""")


#Sound Player class
class Player(object):
#load default settings
    def __init__(self):
    #load source player
        self.source = al.ALuint(0)
        al.alGenSources(1, self.source)
    #disable rolloff factor by default
        al.alSourcef(self.source, al.AL_ROLLOFF_FACTOR, 0)
    #disable source relative by default
        al.alSourcei(self.source, al.AL_SOURCE_RELATIVE,0)
    #capture player state buffer
        self.state = al.ALint(0)
    #set internal variable tracking
        self._volume = 1.0
        self._pitch = 1.0
        self._position = [0,0,0]
        self._rolloff = 1.0
        self._loop = False
        self.queue = []


#set rolloff factor, determines volume based on distance from listener
    def _set_rolloff(self,value):
        self._rolloff = value
        al.alSourcef(self.source, al.AL_ROLLOFF_FACTOR, value)

    def _get_rolloff(self):
        return self._rolloff


#set whether looping or not - true/false 1/0
    def _set_loop(self,lo):
        self._loop = lo
        al.alSourcei(self.source, al.AL_LOOPING, lo)

    def _get_loop(self):
        return self._loop
      

#set player position
    def _set_position(self,pos):
        self._position = pos
        x,y,z = map(int, pos)
        al.alSource3f(self.source, al.AL_POSITION, x, y, z)

    def _get_position(self):
        return self._position
        

#set pitch - 1.5-0.5 float range only
    def _set_pitch(self,pit):
        self._pitch = pit
        al.alSourcef(self.source, al.AL_PITCH, pit)

    def _get_pitch(self):
        return self._pitch

#set volume - 1.0 float range only
    def _set_volume(self,vol):
        self._volume = vol
        al.alSourcef(self.source, al.AL_GAIN, vol)

    def _get_volume(self):
        return self._volume

#queue a sound buffer
    def add(self,sound):
        al.alSourceQueueBuffers(self.source, 1, sound.buf) #self.buf
        self.queue.append(sound)

#remove a sound from the queue (detach & unqueue to properly remove)
    def remove(self):
        if len(self.queue) > 0:
            al.alSourceUnqueueBuffers(self.source, 1, self.queue[0].buf) #self.buf
            al.alSourcei(self.source, al.AL_BUFFER, 0)
            self.queue.pop(0)

#play sound source
    def play(self):
        al.alSourcePlay(self.source)

#get current playing state
    def playing(self):
        al.alGetSourcei(self.source, al.AL_SOURCE_STATE, self.state)
        if self.state.value == al.AL_PLAYING:
            return True
        else:
            return False

#stop playing sound
    def stop(self):
        al.alSourceStop(self.source)

#rewind player
    def rewind(self):
        al.alSourceRewind(self.source)

#pause player
    def pause(self):
        al.alSourcePause(self.source)

#delete sound source
    def delete(self):
        al.alDeleteSources(1, self.source)

#Go straight to a set point in the sound file
    def _set_seek(self,offset):#float 0.0-1.0
        al.alSourcei(self.source,al.AL_BYTE_OFFSET,int(self.queue[0].length * offset))

#returns current buffer length position (IE: 21000), so divide by the buffers self.length
    def _get_seek(self):#returns float 0.0-1.0
        al.alGetSourcei(self.source, al.AL_BYTE_OFFSET, self.state)
        return float(self.state.value)/float(self.queue[0].length)

    rolloff = property(_get_rolloff, _set_rolloff,doc="""get/set rolloff factor""")
    volume = property(_get_volume, _set_volume,doc="""get/set volume""")
    pitch = property(_get_pitch, _set_pitch, doc="""get/set pitch""")
    loop = property(_get_loop, _set_loop, doc="""get/set loop state""")
    position = property(_get_position, _set_position,doc="""get/set position""")
    seek = property(_get_seek, _set_seek, doc="""get/set the current play position""")

if __name__ == '__main__':
    main()
